#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to export an image under different sizes

# (c) Troller0001 2021
#
#   History:
#
#   v0.0: 2021-06-23 First published version

# -------------------------------------------------------------------------

#   Adapted from ofn-export-sizes by ofnuts UNDER THE SAME GPLv3 LICENSE
#   URL: https://sourceforge.net/projects/gimp-tools/files/scripts/
#
#   (c) Ofnuts 2021
#
#   History:
#
#   v0.0: 2021-02-24 First published version
#   v0.1: 2021-02-27 Add non-square shapes

# -------------------------------------------------------------------------

#   Adapted from a template by Herbie
#   URL: http://gimpchat.com/viewtopic.php?f=9&t=18401

# -------------------------------------------------------------------------

#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published
#   by the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This very file is the complete source code to the program.
#
#   If you make and redistribute changes to this code, please mark it
#   in reasonable ways as different from the original version. 
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   The GPL v3 licence is available at: https://www.gnu.org/licenses/gpl-3.0.en.html

import os, json
from datetime import datetime

import gimp, gimpplugin
pdb = gimp.pdb
from gimpenums import *

import gimpui
import pygtk, gtk
pygtk.require('2.0')

# ---------------------------------------------------------------------
# TODO:
# - "Icon Size Presets" -> buildings, techs, civs, etc.
#       - "Custom icon size selection" checkbox
#       - "Promotion Icon Presets" -> Automatically add the blue backgrounds for the 256px (round) and 32px (square) variants
# ---------------------------------------------------------------------

# Tibs Image
tibs = (""
  "GdkP"
  "\0\0\12\350"
  "\2\1\0\1"
  "\0\0\0\\"
  "\0\0\0\36"
  "\0\0\0\36"
  "\177!\36\23\26\25\21\31\31\12\37\37\22\35\35\23\6\12\2\16\13\7\5\4\1"
  ")+\35\14\21\5\12\13\6\12\15\13\3\6\3\0\3\1\27\35\16""03&\2\4\1\30\27"
  "\13AB,\15\13\3\15\13\1\36\31\17\77C//6\35--\26CA!MI+=6\37\36\36\14'*"
  "\30\0\0(%\30#\36\17!%\16##\24\32\33\21\16\17\6$!\33\16\10\7\22\22\13"
  "+2\37\16\21\5\4\3\5\7\10\10\1\2\1\3\4\0\40\"\30\7\4\4\12\7\1A<%0+\36"
  "\7\3\0\32\27\13GN4:C);7%NI1KG0.(\33\6\4\1\4\4\1\0\0$!\20\25\31\11,+\27"
  "//\36\31\33\17\11\13\2\30\35\23&(\35\1\1\1\12\22\13\32\36\26\13\7\11"
  "\2\2\7\1\1\1\2\2\1\23\25\13\26\26\17\20\13\1A=%MJ8\26\22\6\27\26\7PZ"
  "@Y]EVQCGA3-(\23$#\17\10\13\1\1\1\0\0\0\"\37\23\26\30\11)*\31""51\36\33"
  "\33\10\7\10\3\13\15\4%&\30\22\33\20\1\0\1\3\4\4\13\12\7\21\15\16\25\21"
  "\15\21\20\3\"\37\21.'\25.&\17OD,`Z>1,\22""52\37loZpp^:/-&\37\25F@)DA"
  "*$$\25\14\17\14\0\0\21\22\3\22\23\7##\23""73\33""5\177.\31\10\12\3\7"
  "\5\1\35\36\14""26!\27\40\20\2\3\1\16\15\10/(\40$\40\24-(\30""62\40<6"
  "\"F=%WI1_R,QG!rkKllX\40\37\31\37\33\16]Z\77JG,4/\31""3/\36""2-\40\0\0"
  "\30\25\13\24\32\14\35!\20""3/\30aW@21#\3\2\0\20\23\6""0-\32FF+\"\40\22"
  "\32\34\13;<)33\33A\77#;7\32B\77#PI-I@#f]/\177wAvmN\30\24\14!\37\20ic"
  "BPJ-0*\20.(\20""3+\32B;&\0\0""3-#*+\30--\27:4\33SQ5SN2\33\30\12\13\12"
  "\3""0.\32TP0JF)@;&\77""8\40A<\40B=\32AA\35JE&SM/JE$\206\200P|uJ\37\27"
  "\11=5\36ugEG\77\37;1\36+%\23""0*\26""5/\35B8\"\0\0""0.\34\77=';5\35X"
  "T7^[<]_9]U990\36""40\26""76\24A\77\33>7!=3\40=6\36HF'A>$RN3VN-D>\33\204"
  "\200OTP1'!\12wiCZN,+*\20""2/\32\32\27\11(#\25""0,\32C=*\0\0!\33\21+$"
  "\24+$\23;2!HD1QQ9f`CrnHmk\177Aec7WQ)A7\31-#\17""50\35A<#I\77\"QG+d[7"
  "ZP.}r\77XR/\35\22\1o_C1)\23""0)\31\77""8$*&\23(%\22""64%91#\0\0#\36\23"
  "(#\23/&\27""3)\34&\37\23\"\32\22\36\22\14\"\31\15""92\32UO.h]5dW4</\31"
  "9-\32A;\">4\31=1\24KA\"UH'fV+I<\35\40\22\2K\77""3\31\24\3/$\23:2\37="
  ";$QL<55$\30\25\3\0\0(%\23""5/\30;4\34KB+UM2^Y;_N8`Q7MA-\31\16\2""6-\26"
  "h[;\\O2:+\33""6-\32""0(\22,%\22>1\31C7\34J\77\31NC\40#\32\3\34\25\11"
  "\26\24\5""71\30J@$QJ0.(\32('\22PM0\0\0\"\36\12%\37\12!\36\12$\34\6(\36"
  "\15""1,\26""71\26UN%ulHDA$\7\4\1@=#F:\37/%\25'#\13''\14+&\20""9/\21@"
  "5\26""2(\11C9\26E>\35&!\12/-\23:\77\40WW8_\\>dbO\200\177lheG\0\0$!\14"
  "\35\35\13\25\27\10\35\30\15%\33\16\40\34\12),\27%\"\7\77;\31^\\=\23\21"
  "\3\40\27\16""1)\30\177$\35\16\"\34\10!\33\10+#\15""6*\17;/\24/$\13""7"
  "(\14H>\30]Q/CB\40""89\31WW5ijJurUecEb]8\0\0&!\25\32\37\16\37&\21&#\24"
  "\"\34\16\27\25\2%$\21'&\23\31\31\2-,\24\23\26\6\27\22\6)$\26\27\24\7"
  "\23\15\6\32\26\12%\"\12""5+\27/*\15,'\21#\37\5]V+|q@f^8<3\24+$\6HC&`"
  "U5f_:\200[P\0\0\32\31\6\31\34\10\31\33\10\30\30\2\31\24\1\21\26\3($\24"
  "(#\17-,\30\37\37\10\31\37\5""1(\25*&\17\16\23\1\13\13\2\17\17\5\40\31"
  "\6,*\21$)\15\34\35\10\25\20\1UJ\"\213\204LtlAIA!XN/f[=phBmf\77\177wN"
  "\0\0""80\32""1,\25,(\22&!\16,*\26""2.\34""71\36""6.\30<8$21\26""89\31"
  "OG'GB!60\30\31\23\12\5\7\0\20\15\3!\34\14%\35\11-%\27#\31\10SI\40\236"
  "\226fvpEtnFxqNrjJ}yR|rNneE\0\0E>*C9))%\22!\36\13,+\31)!\21/+\30;4\31"
  "73\36.-\23,)\13RJ(_W/d^0-%\17\3\1\0\33\22\13""4\177\"\27L2#D3*%\37\15"
  "\211\202]\253\246\200\232\222qwqNztP\205\202^\204\202YnkCFB%\0\0\21\21"
  "\5\15\7\1\17\12\1\16\17\1\14\21\1\32\32\10>;\37""87\33""52\35\"\35\13"
  "2/\23LE']\\2ql8c^<<3!\32\26\13\22\4\1""7\40\26&!\22=\77)\240\237\200"
  "\252\247\202\232\226q\227\224j\210\205]\202}XspKGD!!\37\5\0\0\4\6\1\11"
  "\15\2\37\34\15""52\34:<\40HG(SK-HE(;;$)&\25)'\16/-\21\\\\6ig8\200\200"
  "L\211\207\\liJ($\25\15\6\3\23\27\14kqY\221\226x\237\236}\233\233z\221"
  "\215h\207\202WwrISL/%\36\11&#\7\0\0\12\12\6#\35\16+'\20.,\23""34\32A"
  "<$OI3QN11/\31,,\24JI.KI(^[4ok>\205\202N\214\213\\\222\222eqnQ52\35""5"
  "2#hbPwsV\215\214k\223\216m[uW{tRmfBD>$:3\31RP-\0\0\15\13\5\26\26\5""0"
  ".\24/0\27%*\17##\20;;'<9#$\"\13IG-TP.A<\27YS+ji9~|J\200\200QtqDhaC9."
  "\31\35\23\5F7,`U\177=xpL\200zV~yYpkJB\77\36""82\27KJ+hc@\0\0\16\24\14"
  "\7\12\2\27\32\12""44#13\31\"%\22!\"\23('\26)*\24QR2ED\37]Y5eb:rqBxwD"
  "ooAoiA[P0=2\30C4!UC1`R5slC\211\204\\uoNID%=8\27C=%9)\26""2\35\11\0\0"
  "\22\25\16\11\12\10\7\2\2\17\23\12\33\35\17$'\32\36!\24\"\40\17\35\35"
  "\16HH0KI-PL2b_\77roHfj\77_Z=H>'[M,h\\9tcA\201nP\222\201a\226\214b\200"
  "wUMG)@:\33A=\40""6,\30-\25\7""1\20\11\0\0\15\20\14\15\15\15\10\14\4\11"
  "\17\7\6\12\6\16\22\12#$\30\37\36\15\40\35\17""53\33:7\32GD*SQ/a^3II)"
  ";6\35LF+kbA~uO\201sR\226\203g\237\214m\206\177[ZU5=:\36ED\"7.\26&\32"
  "\11\"\14\1""3\22\12\0\0\25\26\23\13\15\10\14\23\16\13\17\7\15\15\11\10"
  "\11\5\10\15\11\37\34\20'!\23""67\37""7:\32\77\77\36ED&MP*NQ3^^DrrSrp"
  "N\206\201\\\215\204g\216\204`\212~ZifEPN5<7#E@\40\177PK)MD(/\33\15""9"
  "\35\7\0\0\6\14\10\21\30\23\17\17\16\15\14\11\12\11\3\12\6\11\5\6\4\22"
  "\20\11\37\31\23""54'7=\40GC'UO6OP5WV@abHdeGomH\201}Y|uXymRfaA;=$.-\33"
  "-)\31<5\40MH%piFaR4J-\23\0\0\30\35\27\14\14\12\25\25\24\15\15\11\20\16"
  "\10\23\21\17\22\16\16\17\16\13\11\10\5\21\24\10\27\34\11""77$>7#;8$3"
  "1\34ID,94\33QJ.ID&GA/:</x\177i)-\35\32\27\11\37\37\14\35\34\17\22\21"
  "\1""74\32vmOaO0\0\0\27\33\20\20\23\14\22\27\21\11\15\5\12\16\4\21\22"
  "\17\24\21\21\16\17\13\14\14\11\12\16\6\17\26\14\32\33\20,(\34\35\34\15"
  "\31\31\10\40\35\10\26\25\2$\"\15\"\"\12\37\36\24-8,=I7&*\32&$\22-1\31"
  "#$\17\22\22\6\21\16\2""94!dU:\0\0\16\20\10\26\27\21\11\16\11\17\22\11"
  "\23\26\14\17\20\13\17\13\13\16\15\7\27\32\25\16\21\17\13\23\13\12\14"
  "\3\34\32\16\22\24\10\24\27\13\36\35\16)(\30\32\31\5\"\40\17\26\31\22"
  "CK\77\77E8@A.73\35""2.\26-,\23%&\21\30\27\14\15\14\0@<)\0\37\0\16\20"
  "\17\21\21\16\22\21\16\15\20\4\27\32\15./!\20\17\6\15\16\11\23\26\17\17"
  "\23\15\10\21\7\21\25\14\21\22\11\24\25\14**\34''\24""15\33;9\"85\40\22"
  "\22\3\11\11\2\24\26\14\30\27\11#\36\7""1/\25""21\32$&\17\36\35\20\26"
  "\26\12)&\17\0\0")

# Enums were only introduced in Python 3.4, but this can be used in 2.7 as well
#   URL: https://stackoverflow.com/a/1695250
def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)

# DDS Export Options
class DDS():
    COMPRESSION = enum("NONE", "BC1_DXT1", "BC2_DXT3", "BC3_DXT5", "BC3nm_dxt5NM", "BC4_ATI1_3Dcplus", "BC5_ATI2_3Dc", "RXGB_DXT5", "Alpha_Exponent_DXT5", "YCoCg_DXT5", "YCoCg_scaled_DXT5")
    
    MIPMAPS = enum("NO_MIPMAPS", "GENERATE_MIPMAPS", "USE_EXISTING_MIPMAPS")
    
    SAVE_TYPE = enum("SELECTED_LAYER", "CUBE_MAP", "VOLUME_MAP", "TEXTURE_ARRAY")
    
    FORMAT = enum("DEFAULT", "RGB8", "RGBA8", "BGR8", "ABGR8", "R5G6B5", "RGBA4", "RGB5A1", "RGB10A2", "R3G3B2", "A8", "L8", "L8A8", "AExp", "YCoCg")
    
    # index of transparent colour (-1 to disable)
    TRANSPARENT_INDEX = enum(value=staticmethod(lambda val: val))
    
    MIPMAP_FILTER = enum("DEFAULT", "NEAREST", "BOX", "TRIANGLE", "QUADRATIC", "BSPLINE", "MITCHELL", "LANCZOS", "KAISER")
    
    MIPMAP_WRAP_MODE = enum("DEFAULT", "MIRROR", "REPEAT", "CLAMP")
    
    MIPMAP_APPLY_GAMMA_CORRECT = enum("NO", "YES")
    
    MIPMAP_USE_SRGB_COLORSPACE = enum("NO", "YES")
    
    # (0, 100)
    MIPMAP_GAMMA = enum(value=staticmethod(lambda val: val))
    
    COMPRESSION_PERCEPTUAL_ERROR_METRIC = enum("NO", "YES")
    
    MIPMAP_PRESERVE_ALPHA_TEST_COVERAGE = enum("NO", "YES")
    
    # (0, 1)
    MIPMAP_ALPHA_TEST_THRESHOLD = enum(value=staticmethod(lambda val: val))
    

##########################################################################################
# User interface class.
class IconExportUserInterface(gtk.Window):
    default_base_filename = "civ5_icon_sheet" # Standard base filename used when exporting an unsaved file
    default_directory = "." # Export icons to this location by default
    available_icon_sizes = [16, 22, 24, 32, 45, 48, 64, 80, 128, 214, 256] # Icon sizes that can be exported to
    
    iconatlas_exports = [
        {'name': "XML", 'extension': ".xml", 'method': "write_iconatlas_xml"},
        {'name': "SQL", 'extension': ".sql", 'method': "write_iconatlas_sql"},
    ]
    
    # Unused
    icon_sizes_per_type = {
        "Buildings": [45, 64, 80, 128, 256],
        "Civs": [32, 45, 64, 80, 128, 256],
    }

    #=======================================================================================
    # The constructor that creates the window.
    def __init__(self, fImage):
        #
        #-------------------------------------------------------------------------------------
        # Basic Error Handling
        
        # Should have dds export options available
        if not pdb.query('file-dds-save'):
            self.show_error(title="ERROR: DDS Plugin not installed!", msg="The GIMP DDS Plugin (by Shawn Kirst) was not installed, but this is needed to export civ5 icons.\n\nNote that this DDS plugin became native to GIMP 2.10.10, so upgrading GIMP to this version also suffices.\nCurrent GIMP version: " + pdb.gimp_version(), show_ribbon_msg=False)
            return
        
        # Plugin is not compatible with GIMP 3 (which uses GTK 3 and Python 3.x)
        if int(pdb.gimp_version().split('.')[0]) >= 3:
            self.show_error(title="ERROR: Not Compatible with GIMP 3+!", msg="The civ5 icon export plugin is not (yet) compatible with GIMP 3+. Please revert to GIMP 2 (version 2.10.10 or later)\n\nCurrent GIMP version: " + pdb.gimp_version(), show_ribbon_msg=False)
            return
    
        #
        #-------------------------------------------------------------------------------------
        # Main point of reference
        self.image = fImage
        
        #
        #-------------------------------------------------------------------------------------
        # Title, size, position, .. of the window.
        gtk.Window.__init__(self, gtk.WINDOW_TOPLEVEL)
        self.set_title('Export Civ5 Icons')
        self.winWidth = 400
        self.winHeight = 300
        self.set_default_size(self.winWidth, self.winHeight)
        self.set_size_request(self.winWidth, -1)
        self.set_position(gtk.WIN_POS_CENTER)
        self.set_border_width(10)
        self.set_resizable(False)
        
        #
        #-------------------------------------------------------------------------------------
        # Define container (table) for the widget elements.
        self.layTab = gtk.VBox(spacing=5)
        self.layTab.set_border_width(5)
        alignment = gtk.Alignment(xscale=1.0)
        alignment.add(self.layTab)
        self.add(alignment)
        
        #
        #-------------------------------------------------------------------------------------
        # Previous User Input
        
        try:
            self.config = json.loads(gimp.get_data("TRL-export-civ5"))
        except gimp.error as e:
            # No previous config values found (first time running this plugin since starting up GIMP)
            self.config = {
                "current_icon_size": 256,
                "filename": self.default_base_filename,
                "directory": self.default_directory,
                "icon_sizes": filter(lambda sz: sz in [32, 45, 64, 80, 128, 256], self.available_icon_sizes), # Enabled icon sizes
                "compress": True,
                "upscale_to_multiple_of_four": True,
                "resize": False,
                "icons_per_row": 8,
                "icons_per_column": 8,
                "iconatlas": 0, # Do not generate
            }
        
        #
        #-------------------------------------------------------------------------------------
        # User Input
        
        ##########################
        # Current Icon Size
        self.curIconSizeButton = gtk.SpinButton()
        self.curIconSizeButton.set_numeric(True)
        self.curIconSizeButton.set_range(1, min(self.image.width, self.image.height))
        self.curIconSizeButton.set_increments(1, 50)
        default_icon_size = self.config["current_icon_size"]
        if self.image.width < default_icon_size or self.image.height < default_icon_size:
            default_icon_size = min(self.image.width, self.image.height)
        self.curIconSizeButton.set_value(default_icon_size)
        self.add_vbox_item_2col(self.layTab, left_item=gtk.Label(str="Current Icon Size"), right_item=self.curIconSizeButton, tooltip_text="The current icon size is used to determine how to scale the icon sheet. It is recommended to make icons at least the size you're exporting to, as otherwise the icon will need to be upscaled, which has reduced quality.")
        
        ##########################
        # Export filename & Directory
        imageName = self.config["filename"]
        imageDirectory = self.config["directory"]
        if self.image.filename:
            # If working with a saved file, default to that file's name and directory
            if imageName == self.default_base_filename:
                # Unless the filename was overridden before
                imageName, _ = os.path.splitext(os.path.basename(self.image.filename))
            
            if imageDirectory == self.default_directory:
                # Unless the output directory was overridden before
                imageDirectory = os.path.dirname(self.image.filename)
        
        ##########################
        # Filename input
        self.filenameInput = gtk.Entry()
        self.filenameInput.set_text(imageName)
        self.add_vbox_item_2col(self.layTab, left_item=gtk.Label(str="Export Name"), right_item=self.filenameInput, tooltip_text="A prefix to use for the exported sheets' filenames.\nExample: An export name of \"my_icon_sheet\" results in filenames of the form: \"my_icon_sheet_256.dds\", \"my_icon_sheet_45.dds\", etc.")
        
        ##########################
        # Directory Input
        self.directoryInput = gtk.FileChooserButton(gtk.FileChooserDialog(title="Select Export Directory", parent=self, action=gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER, buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT, gtk.STOCK_OK, gtk.RESPONSE_ACCEPT)))
        self.directoryInput.set_filename(imageDirectory)
        # TODO: add_shortcut_folder?
        self.add_vbox_item_2col(self.layTab, left_item=gtk.Label(str="Export Directory"), right_item=self.directoryInput)
        
        ##########################
        # Lay out checkboxes for icon sizes
        numIconSizeColumns = 3
        numIconSizeRows = len(self.available_icon_sizes) // numIconSizeColumns + 1
        checkboxTable = gtk.Table(rows=numIconSizeRows, columns=numIconSizeColumns, homogeneous=True)
        self.icon_sizes_inputs = {}
        for i, icon_size in enumerate(self.available_icon_sizes):
            # Create a checkbox
            icon_size_checkbox = gtk.CheckButton(label=str(icon_size))
            icon_size_checkbox.set_active(icon_size in self.config["icon_sizes"])
            self.icon_sizes_inputs[icon_size] = icon_size_checkbox
            
            # Align it
            alignment = gtk.Alignment(xalign=0.5, yscale=1.0)
            alignment.add(icon_size_checkbox)
            
            # Determine position in grid
            col, row = divmod(i, numIconSizeRows) # = (i // numrows, i % numrows)
            checkboxTable.attach(alignment, col, col + 1, row, row + 1)
        
        iconframe = gtk.Frame(label="Icon Sizes")
        iconframe.set_label_align(0.5, 0.5)
        iconframe.add(checkboxTable)
        self.layTab.add(iconframe)
        
        ##########################
        # Horizontal bar
        self.layTab.add(gtk.HSeparator())
        
        ####################################################
        # Advanced Options
        expander = gtk.Expander(label="Advanced Options")
        advOptsVBox = gtk.VBox(spacing=5)
        expander.add(advOptsVBox)
        self.layTab.add(expander)

        ##########################
        # Generate IconAtlas
        self.genIconAtlasDropdown = gtk.combo_box_new_text()
        self.genIconAtlasDropdown.append_text('Do not Generate')
        for iconatlas_export in self.iconatlas_exports:
            self.genIconAtlasDropdown.append_text("Generate " + iconatlas_export['name'])
        self.genIconAtlasDropdown.set_active(self.config["iconatlas"])
        self.add_vbox_item_2col(advOptsVBox, left_item=gtk.Label(str="Generate IconAtlas"), right_item=self.genIconAtlasDropdown)
        
        ##########################
        # Compression
        compressionVBox = gtk.VBox(spacing=5)
        compressionVBox.set_border_width(5)
        compressionframe = gtk.Frame(label="Compression")
        compressionframe.set_label_align(0.5, 0.5)
        compressionframe.add(compressionVBox)
        advOptsVBox.add(compressionframe)
        
        #############
        # Compression enabled checkbox
        self.compressCheckbox = gtk.CheckButton(label="Compress Icon Sheets (BC2 / DXT3)")
        self.compressCheckbox.set_active(self.config["compress"])
        self.compressCheckbox.set_tooltip_text("Compression lowers file size, but also reduces icon quality.")
        
        # Disable "multiple-of-four"-checkbox if compression is disabled
        def on_compress_toggle(compressCheckbox):
            self.upscaleMultipleFourCheckbox.set_sensitive(compressCheckbox.get_active())
            
        self.compressCheckbox.connect("clicked", on_compress_toggle)
        compressionVBox.add(self.compressCheckbox)
        
        #############
        # Upscale to multiple of 4
        self.upscaleMultipleFourCheckbox = gtk.CheckButton(label="Upscale sheet dimensions to multiple of 4")
        self.upscaleMultipleFourCheckbox.set_active(self.config["upscale_to_multiple_of_four"])
        self.upscaleMultipleFourCheckbox.set_tooltip_text("Compression can only be performed if the icon sheet dimensions are a multiple of four.\n- If enabled, resizes the icon sheet up to the nearest multiple of four.\n- If disabled, files with a width and/or height that are not a multiple of four will not be compressed.")
        self.upscaleMultipleFourCheckbox.set_sensitive(self.compressCheckbox.get_active())
        compressionVBox.add(self.upscaleMultipleFourCheckbox)
        
        ##########################
        # Resize Icon Sheet
        resizeframe = gtk.Frame(label="Icon Sheet Size")
        resizeframe.set_label_align(0.5, 0.5)
        resizeVBox = gtk.VBox(spacing=5)
        resizeVBox.set_border_width(5)
        resizeframe.add(resizeVBox)
        advOptsVBox.add(resizeframe)
        
        #############
        # Checkbox
        self.resizeCheckbox = gtk.CheckButton(label="Resize Icon Sheet")
        self.resizeCheckbox.set_active(self.config["resize"])
        self.resizeCheckbox.set_tooltip_text("Cuts the icon sheet to fit the given amount of icons.")
        
        # Disable "icons-per-row/column" inputs if icon sheet resizing is disabled
        def on_resize_toggle(resizeCheckbox):
            self.iconsPerRowInput.set_sensitive(resizeCheckbox.get_active())
            self.iconsPerColumnInput.set_sensitive(resizeCheckbox.get_active())
            
        self.resizeCheckbox.connect("clicked", on_resize_toggle)
        self.add_vbox_item_2col(resizeVBox, right_item=self.resizeCheckbox)
        
        #############
        # Icons per row
        self.iconsPerRowInput = gtk.SpinButton()
        self.iconsPerRowInput.set_numeric(True)
        self.iconsPerRowInput.set_range(1, 8)
        self.iconsPerRowInput.set_increments(1, 2)
        icons_per_row = self.config["icons_per_row"]
        self.iconsPerRowInput.set_value(icons_per_row)
        self.iconsPerRowInput.set_sensitive(self.resizeCheckbox.get_active())
        self.add_vbox_item_2col(resizeVBox, left_item=gtk.Label(str="Icons per Row"), right_item=self.iconsPerRowInput, tooltip_text="Number of icons in a single row.")
        
        #############
        # Icons per column
        self.iconsPerColumnInput = gtk.SpinButton()
        self.iconsPerColumnInput.set_numeric(True)
        self.iconsPerColumnInput.set_range(1, 8)
        self.iconsPerColumnInput.set_increments(1, 2)
        icons_per_column = self.config["icons_per_column"]
        self.iconsPerColumnInput.set_value(icons_per_column)
        self.iconsPerColumnInput.set_sensitive(self.resizeCheckbox.get_active())
        self.add_vbox_item_2col(resizeVBox, left_item=gtk.Label(str="Icons per Column"), right_item=self.iconsPerColumnInput, tooltip_text="Number of icons in a single column.")
        
        ##########################
        # Horizontal bar
        self.layTab.add(gtk.HSeparator())
        
        ##########################
        # Hidden Layer Warning
        hiddenLayerWarningLabel = gtk.Label()
        al = gtk.Alignment()
        hiddenLayerWarningLabel.set_markup("<small><i>Note: Hidden Layers will be ignored when generating icon sheets.</i></small>")
        al.add(hiddenLayerWarningLabel)
        self.layTab.add(al)
        
        ##########################
        # Ok Button
        self.okButton = gtk.Button("Generate")
        self.okButton.set_relief(gtk.RELIEF_NORMAL)
        self.okButton.connect("clicked", self.on_ok)
        self.add_vbox_item_2col(self.layTab, right_item=self.okButton)
        
        #-------------------------------------------------------------------------------------
        # Close the user interface with the X button.
        self.connect("destroy", self.on_exit) 
    
    # Add a pair of items in a 2-column (even spacing) format to a given VBox, with a given tooltip.
    # This allows, for instance, to add a (Label, Entry) pair.
    def add_vbox_item_2col(self, vbox, left_item=None, right_item=None, tooltip_text=None, homogeneous=True):
        # Right-align the left column
        lalignment = gtk.Alignment(xalign=1.0, yalign=0.5)
        if left_item:
            lalignment.add(left_item)
        
        # Do not align the right column
        ralignment = gtk.Alignment(yalign=0.5, xscale=1.0)
        if right_item:
            ralignment.add(right_item)
        
        # Add both to an hbox, and attach the hbxo to the given vbox
        hbox = gtk.HBox(homogeneous=homogeneous, spacing=10)
        hbox.add(lalignment)
        hbox.add(ralignment)
        hbox.set_tooltip_text(tooltip_text)
        vbox.add(hbox)
        
    #
    #=======================================================================================
    # Display an error
    def show_error(self, title, msg, show_ribbon_msg=True):
        if show_ribbon_msg:
            pdb.gimp_message(msg)
        errorDialog = gtk.MessageDialog(parent=self, flags=gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT, type=gtk.MESSAGE_ERROR, buttons=gtk.BUTTONS_OK, message_format=msg)
        errorDialog.set_title(title)
        errorDialog.show()
        errorDialog.run()
        errorDialog.destroy()
    #
    #=======================================================================================
    # Method of displaying the window and staying in the event queue.
    def showInterface(self):
        self.show_all()    # Now display the window built in the constructor
        gtk.main()          # Event queue after displaying the main window
    #
    #=======================================================================================
    # Event in response to the key being pressed on "OK"
    def on_ok(self, widget, data=None):
        ##########################
        # Get selected icon sizes
        export_icon_sizes = []
        for icon_size, icon_size_checkbox in self.icon_sizes_inputs.items():
            if icon_size_checkbox.get_active():
                export_icon_sizes.append(icon_size)
        
        #############
        # Verify at least one icon size is selected
        if not export_icon_sizes:
            self.show_error(title="ERROR: No Icon Size selected!", msg="Should select at least one icon size to export to.")
            return
            
        #############
        # Verify that a filename was passed
        if not self.filenameInput.get_text():
            self.show_error(title="ERROR: No filename given!", msg="The export name cannot be empty.")
            return
        
        ##########################
        # Calculate sizes & Read settings
        width = int(self.image.width)
        height = int(self.image.height)
        current_icon_size = int(self.curIconSizeButton.get_value())
        icons_per_row = int(width // current_icon_size)
        icons_per_column = int(height // current_icon_size)
        base_filename = self.filenameInput.get_text()
        export_directory = self.directoryInput.get_filename()
        compress_enabled = self.compressCheckbox.get_active()
        upscale_to_multiple_of_four_enabled = self.upscaleMultipleFourCheckbox.get_active()
        resize_enabled = self.resizeCheckbox.get_active()
        resize_icons_per_row = int(self.iconsPerRowInput.get_value())
        resize_icons_per_column = int(self.iconsPerColumnInput.get_value())
        iconatlas = self.genIconAtlasDropdown.get_active()
        
        ##########################
        # Verify that the icon sheet resize dimensions actually fit on the current sheet
        if resize_enabled and (resize_icons_per_row > icons_per_row or resize_icons_per_column > icons_per_column):
            self.show_error(title="ERROR: Icon Sheet resize exceeds image dimensions!", msg="The number of icons per row (" + str(resize_icons_per_row) + ") or column (" + str(resize_icons_per_column) + ") exceeds the amount that can fit on the sheet. Please reduce the number(s) or disable icon sheet resizing.\n\nMaximum icons (based on current image dimensions):\n* Per Row: " + str(icons_per_row) + "\n* Per Column: " + str(icons_per_column), show_ribbon_msg=False)
            return
        
        ##########################
        # Store export settings
        self.config = {
            "current_icon_size": current_icon_size,
            "filename": base_filename,
            "directory": export_directory,
            "icon_sizes": export_icon_sizes,
            "compress": compress_enabled,
            "upscale_to_multiple_of_four": upscale_to_multiple_of_four_enabled,
            "resize": resize_enabled,
            "icons_per_row": resize_icons_per_row,
            "icons_per_column": resize_icons_per_column,
            "iconatlas": iconatlas,
        }
        gimp.set_data("TRL-export-civ5", json.dumps(self.config))
        
        ##########################
        # Enable cutting of the icon sheet
        if resize_enabled:
            icons_per_row = resize_icons_per_row
            icons_per_column = resize_icons_per_column
        
        ##########################
        # The plugin ends up in the same state as it was started in.
        # Although it does perform operations that can (in theory) be undone, this is unwanted behaviour, so we prevent "undo/redo" instead
        self.image.undo_freeze()
        
        # Were there any changes before the plugin did stuff?
        was_dirty = self.image.dirty
        
        ##########################
        # Use Cubic Interpolation for scaling
        old_interpolation = pdb.gimp_context_get_interpolation()
        pdb.gimp_context_set_interpolation(INTERPOLATION_CUBIC)
        
        ##########################
        # Export a .dds file for each icon size
        for icon_size in export_icon_sizes:
            ##########################
            # Generate filename
            export_filename = base_filename + "_" + str(icon_size) + ".dds"
            export_path = os.path.join(export_directory, export_filename)
            
            ##########################
            # Create a new layer by merging all visible layers, and add it to the image
            saveLayer = pdb.gimp_layer_new_from_visible(self.image, self.image, export_filename)
            self.image.insert_layer(saveLayer, position=0)
            
            ##########################
            # Cut off extra unused spacing
            #   Assumes that the icons are located in the top-left corner of the sheet
            #   This should only be needed if non-standard dimensions are used (e.g., a single 45px icon is located in a 64*64px sheet)
            saveLayer.resize(current_icon_size*icons_per_row, current_icon_size*icons_per_column)
            
            ##########################
            # Scale the layer to the requested size
            new_width = icon_size*icons_per_row
            new_height = icon_size*icons_per_column
            saveLayer.scale(new_width, new_height, False)
            
            ##########################
            # Add an alpha mask based on the layer's alpha channel
            saveLayerMask = saveLayer.create_mask(ADD_ALPHA_MASK)
            saveLayer.add_mask(saveLayerMask)
            
            ##########################
            # Change dds export settings
            dds_format = DDS.FORMAT.RGBA8
            dds_compression = DDS.COMPRESSION.NONE
            if compress_enabled:
                do_compress = False
                next_multiple_of_four_width = ((new_width + 3) / 4) * 4 
                next_multiple_of_four_height = ((new_height + 3) / 4) * 4 
                
                if new_width != next_multiple_of_four_width or new_height != next_multiple_of_four_height:
                    if upscale_to_multiple_of_four_enabled:
                        saveLayer.resize(next_multiple_of_four_width, next_multiple_of_four_height)
                        do_compress = True
                else:
                    do_compress = True
                
                
                if do_compress:
                    dds_format = DDS.FORMAT.DEFAULT
                    dds_compression = DDS.COMPRESSION.BC2_DXT3
            
            ##########################
            # Export as a dds file
            pdb.file_dds_save(self.image, saveLayer, export_path, export_path,
                dds_compression,
                DDS.MIPMAPS.NO_MIPMAPS,
                DDS.SAVE_TYPE.SELECTED_LAYER,
                dds_format,
                DDS.TRANSPARENT_INDEX.value(0),
                DDS.MIPMAP_FILTER.DEFAULT,
                DDS.MIPMAP_WRAP_MODE.DEFAULT,
                DDS.MIPMAP_APPLY_GAMMA_CORRECT.NO,
                DDS.MIPMAP_USE_SRGB_COLORSPACE.NO,
                DDS.MIPMAP_GAMMA.value(2.2),
                DDS.COMPRESSION_PERCEPTUAL_ERROR_METRIC.NO,
                DDS.MIPMAP_PRESERVE_ALPHA_TEST_COVERAGE.NO,
                DDS.MIPMAP_ALPHA_TEST_THRESHOLD.value(0.5)
            )
            
            ##########################
            # Remove the temporary layer
            self.image.remove_layer(saveLayer)
        
        ##########################
        # Restore old interpolation settings        
        pdb.gimp_context_set_interpolation(old_interpolation)
        
        ##########################
        # Generate IconAtlas
        if iconatlas > 0:
            export_filename = "IconAtlas_" + base_filename + self.iconatlas_exports[iconatlas - 1]['extension']
            export_path = os.path.join(export_directory, export_filename)
            with open(export_path, 'w') as fp:
                getattr(self, self.iconatlas_exports[iconatlas - 1]['method'])(fp, base_filename, export_icon_sizes, icons_per_row, icons_per_column)
        
        ##########################
        # If the image state was clean before, it is still clean now
        if not was_dirty:
            self.image.clean_all()
            
        # Restore "undo/redo" behaviour; the user can still undo actions it made prior to running the plugin
        self.image.undo_thaw()
        
        # ##########################
        # Show completion dialog
        doneDialog = gtk.MessageDialog(parent=None, flags=gtk.DIALOG_MODAL, type=gtk.MESSAGE_INFO, buttons=gtk.BUTTONS_CLOSE, message_format="Icons were exported successfully!")
        doneDialog.set_title("Icon Export Complete!")
        doneDialog.format_secondary_text("Your icon sheets (and IconAtlas) were exported to the following folder:")
        
        # Export Location label
        exportLocationLabel = gtk.Label(str=str(export_directory))
        exportLocationLabel.set_line_wrap_mode(gtk.WRAP_WORD_CHAR)
        exportLocationLabel.set_line_wrap(True)
        exportLocationLabel.set_selectable(True)
        doneDialog.get_message_area().add(exportLocationLabel)
        
        # Separator
        doneDialog.get_message_area().add(gtk.HSeparator())
        
        # Mod Properties tip (VFS=true)
        modPropertiesLabel = gtk.Label(str="Do not forget to set the images to VFS=true, and add OnModActivated->UpdateDatabase for your IconAtlas.")
        modPropertiesLabel.set_line_wrap(True)
        modPropertiesLabel.set_selectable(True)
        doneDialog.get_message_area().add(modPropertiesLabel)
        
        #############        
        # Tibs image
        tibsImage = gtk.Image()
        tibsPixbuf = gtk.gdk.pixbuf_new_from_inline(5860, tibs, False)
        # pixbuf = gtk.gdk.pixbuf_new_from_file("C:/Users/<USER>/Pictures/tibs.png")
        tibsPixbuf = tibsPixbuf.scale_simple(60, 60, gtk.gdk.INTERP_NEAREST)
        tibsImage.set_from_pixbuf(tibsPixbuf)
        
        # Align & Enable Tibs
        alignment = gtk.Alignment(xalign=0.5, yalign=0.0)
        alignment.add(tibsImage)
        alignment.set_padding(8, 0, 2, 2)
        doneDialog.set_image(alignment)
        
        #############
        # Show the dialog
        doneDialog.set_position(gtk.WIN_POS_CENTER)
        doneDialog.show_all()
        self.hide()
        doneDialog.run()
        doneDialog.destroy()
        
        ##########################        
        # Call the exit function
        self.on_exit(widget, data)
        
    ##########################
    # Event in response to the key being pressed on "X".
    def on_exit(self, widget, data=None):
        gtk.main_quit()         # End of the event queue that started with "gtk.main()"
        
    def write_iconatlas_xml(self, fp, base_filename, icon_sizes, icons_per_row, icons_per_column):
        fp.write('<?xml version="1.0" encoding="utf-8"?>\n')
        fp.write("<!-- Generated by TRL-export-civ5 on " + datetime.now().strftime('%Y-%m-%d %H:%M:%S') + " -->\n")
        fp.write("<GameData>\n")
        fp.write("    <IconTextureAtlases>\n")
        for icon_size in icon_sizes:
            fp.write("        <Row>\n")
            fp.write("            <Atlas>" + base_filename.upper() + "</Atlas>\n")
            fp.write("            <IconSize>" + str(icon_size) + "</IconSize>\n")
            fp.write("            <Filename>" + base_filename + "_" + str(icon_size) + ".dds</Filename>\n")
            fp.write("            <IconsPerRow>" + str(icons_per_row) + "</IconsPerRow>\n")
            fp.write("            <IconsPerColumn>" + str(icons_per_column) + "</IconsPerColumn>\n")
            fp.write("        </Row>\n")
        fp.write("    </IconTextureAtlases>\n")
        fp.write("</GameData>\n")

    
    def write_iconatlas_sql(self, fp, base_filename, icon_sizes, icons_per_row, icons_per_column):
        fp.write("-- Generated by TRL-export-civ5 on " + datetime.now().strftime('%Y-%m-%d %H:%M:%S' + "\n"))
        fp.write("INSERT INTO IconTextureAtlases (Atlas, IconSize, Filename, IconsPerRow, IconsPerColumn)\n")
        fp.write("VALUES                  ")
        vals = []
        for icon_size in icon_sizes:
            vals.append("('{}', {}, '{}', {}, {})".format(base_filename.upper(), icon_size, base_filename + "_" + str(icon_size) + ".dds", icons_per_row, icons_per_column))
        fp.write(",\n                        ".join(vals))
        fp.write(";\n")

##########################################################################################
# Plugin class.
class plugin_main(gimpplugin.plugin):
    pluginName = "TRL_export_civ5"
    blurb="Export Civ5 Icons"
    help="Export Civ5 Icons"
    author='Troller0001'
    copyright = "(c) " + author
    date="JUN 2021"
    menuTitle="Export Civ5 Icons..."
    exportMenu="<Image>/File/Export/"

    #
    #=======================================================================================
    # Usually you do not redefine the following three methods in your own plugin,
    # since these are already set appropriately in the upper class. But it works.
    def start(self) :
        gimp.main(self.init, self.quit, self.query, self._run)
    def init(self):
        pass
    def quit(self):
        pass
    #
    #=======================================================================================
    # Method "query(self)" is called when the plugin is recognized as new/changed.
    def query(self):
        gimp.install_procedure(
            self.pluginName,         # Name
            self.blurb,              # Blurb
            self.help,               # Help
            self.author,             # Author
            self.copyright,          # Copyright
            self.date,               # Date
            self.menuTitle,          # Menuname
            '*',                     # Imagetypes
            PLUGIN,
            [   ( # Run mode (32-bit Integer)
                  PDB_INT32,        # A plugin can be started in the following modes:
                  "run_mode",       #   RUN_INTERACTIVE (0), RUN_NONINTERACTIVE (1)
                  "Run mode"        #   and RUN_WITH_LAST_VALS (2)
                ),
                ( # Picture/Image (Image ID)
                  PDB_IMAGE,       # Every opened picture has a single active drawing area,
                  "image",             # the "Drawable".
                  "Input image"
                ),
                ( # Drawable (Drawable ID)
                  PDB_DRAWABLE,  # A "Drawable" is a GIMP concept that contains layers,
                  "drawable",          # but also various other objects such as channels,
                  "Input drawable"  # layer masks and the selection mask.
                )
            ],
            []
        )
        gimp.menu_register(self.pluginName, self.exportMenu)
    #
    #======================================================================================
    # The method "run(self)" realizes the actual plugin.
    def TRL_export_civ5(self, fRunMode, fImage, fDrawable):
        if (fRunMode == RUN_NONINTERACTIVE):
            return

        userInterface = IconExportUserInterface(fImage)
        # GTK: Show surface and wait until "gtk_main_quit()" is called.
        userInterface.showInterface()
           
if (__name__ == '__main__'):
    plugin_main().start()
